import random

import numpy as np
import pytest
import tensorflow as tf

from tests.conftest import sample_state, sample_action


def test_deceleration(navigation, batch_size):
    center = navigation.deceleration["center"]
    decay = navigation.deceleration["decay"]

    state1 = sample_state(navigation)
    state2 = sample_state(navigation, batch_size)

    for state in [state1, state2]:
        batch_size = tf.shape(state)[0]

        lambda_ = navigation._deceleration(state)
        lambda_ = navigation._deceleration(state)

        assert lambda_.shape == [batch_size,]

        for i in range(batch_size):
            state_ = state[i]
            lambda_i = lambda_[i]

            expected = 1.0
            for c, d in zip(center, decay):
                delta = np.squeeze(state_ - c)
                distance = np.sqrt(np.sum(delta ** 2, axis=-1))
                expected *= 2 / (1.0 + np.exp(-d * distance)) - 1.0

            assert isinstance(lambda_i, tf.Tensor)
            assert lambda_i.shape == []
            assert lambda_i.dtype == tf.float32
            assert tf.reduce_all(tf.abs(lambda_i - expected) < 1e-4)

    fn = navigation._deceleration
    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_transition(navigation, batch_size):
    state1 = sample_state(navigation)
    action1 = sample_action(navigation)

    state2 = sample_state(navigation, batch_size)
    action2 = sample_action(navigation, batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        lambda_ = navigation._deceleration(state)

        next_state = navigation.transition(state, action)
        assert next_state.shape == state.shape

        for i in range(batch_size):
            assert tf.reduce_all(
                next_state[i] == state[i] + lambda_[i] * action[i])

    fn = navigation.transition

    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)

    cfn3 = fn.get_concrete_function(state1, action1)
    cfn4 = fn.get_concrete_function(state2, action2)

    assert cfn1 == cfn2 == cfn3 == cfn4


def test_linear_transition(navigation, batch_size):
    goal = navigation.goal
    decay = navigation.deceleration["decay"]
    center = navigation.deceleration["center"]

    state1 = sample_state(navigation)
    action1 = sample_action(navigation)
    state2 = sample_state(navigation, batch_size)
    action2 = sample_action(navigation, batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        lambda_ = navigation._deceleration(state)

        model = navigation.get_linear_transition(state, action)
        f = model.f
        f_x = model.f_x
        f_u = model.f_u

        assert f.shape == state.shape

        assert tf.reduce_all(f == navigation.transition(state, action))

        assert f_u.shape == [batch_size] + [navigation.action_size] * 2
        assert tf.reduce_all(f_u == tf.reshape(lambda_, [-1, 1, 1]) * tf.eye(2))
        assert f_x.shape == [batch_size] + [navigation.state_size] * 2

        for i in range(batch_size):
            state_ = state[i]
            action_ = action[i]

            lambda_x = tf.zeros_like(state_)
            for c, d in zip(center, decay):
                dist = tf.norm(state_ - c)
                l = 2 / (1.0 + tf.exp(-d * dist)) - 1.0

                h = 2 * d * tf.exp(-d * dist) / ((1 + tf.exp(-d * dist)) ** 2)
                lambda_x += h * (state_ - c) / dist * lambda_[i] / l

            expected_f_x = (tf.eye(navigation.state_size)
                            + tf.tensordot(tf.squeeze(action_),
                                           tf.squeeze(lambda_x), axes=0))

            assert tf.reduce_all(tf.abs(f_x[i] - expected_f_x) < 1e-4)

    fn = navigation.get_linear_transition
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_cost(navigation, batch_size):
    goal = navigation.goal

    state1 = sample_state(navigation)
    action1 = sample_action(navigation)
    state2 = sample_state(navigation, batch_size=batch_size)
    action2 = sample_action(navigation, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        cost = navigation.cost(state, action)

        assert isinstance(cost, tf.Tensor)
        assert cost.shape == [batch_size,]
        assert tf.reduce_all(cost >= 0.0)

        cost2 = navigation.cost(tf.expand_dims(goal, axis=0), action)
        assert tf.reduce_all(cost2 == 0.0)

    fn = navigation.cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_quadratic_cost(navigation, batch_size):
    goal = navigation.goal

    state1 = sample_state(navigation)
    action1 = sample_action(navigation)
    state2 = sample_state(navigation, batch_size=batch_size)
    action2 = sample_action(navigation, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]
        state_size = navigation.state_size
        action_size = navigation.action_size

        model = navigation.get_quadratic_cost(state, action)
        l = model.l
        l_x = model.l_x
        l_u = model.l_u
        l_xx = model.l_xx
        l_uu = model.l_uu
        l_xu = model.l_xu
        l_ux = model.l_ux

        assert l.shape == [batch_size,]
        assert tf.reduce_all(l == navigation.cost(state, action))

        assert tf.reduce_all(l_x == 2 * (state - goal))

        assert tf.reduce_all(l_u == tf.zeros_like(action))

        assert tf.reduce_all(l_xx == tf.reshape(
            tf.tile(2 * tf.eye(state_size), [batch_size, 1]),
            [batch_size, state_size, state_size]))

        assert tf.reduce_all(l_uu == tf.zeros([batch_size, action_size, action_size]))

        assert tf.reduce_all(l_ux == tf.zeros([batch_size, state_size, action_size]))

        assert tf.reduce_all(l_xu == tf.zeros([batch_size, action_size, state_size]))

    fn = navigation.get_quadratic_cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_quadratic_final_cost(navigation):
    goal = navigation.goal

    state = navigation.obs_space.sample()

    model = navigation.get_quadratic_final_cost(state)
    l = model.l
    l_x = model.l_x
    l_xx = model.l_xx

    assert l.shape == []
    assert tf.reduce_all(l == navigation.final_cost(state))

    assert tf.reduce_all(l_x == 2 * (state - goal))
    assert tf.reduce_all(l_xx == 2 * tf.eye(navigation.state_size))




import gym
import numpy as np
import tensorflow as tf

from tfmpc.envs.diffenv import DiffEnv
from tfmpc.envs.gymenv import GymEnv


class Navigation(DiffEnv, GymEnv):

    def __init__(self, goal, deceleration, low, high):
        super().__init__()

        self.goal = goal
        self.deceleration = deceleration

        self.obs_space = gym.spaces.Box(
            shape=goal.shape,
            low=-np.inf,
            high=np.inf
        )

        self.action_space = gym.spaces.Box(
            low=np.array(low, dtype=np.float32),
            high=np.array(high, dtype=np.float32),
        )

    @property
    def action_size(self):
        return self.state_size

    @property
    def state_size(self):
        return self.goal.shape[0]

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def transition(self, state, action):
        lambda_ = self._deceleration(state)
        lambda_ = tf.reshape(lambda_, [-1, 1, 1])
        next_state = state + lambda_ * action
        return next_state

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def cost(self, state, action):
        del action
        state = tf.squeeze(state, axis=-1)
        goal = tf.squeeze(self.goal)
        return tf.reduce_sum((state - goal) ** 2, axis=-1)

    @tf.function
    def final_cost(self, state):
        state = tf.squeeze(state, axis=-1)
        goal = tf.squeeze(self.goal)
        return tf.reduce_sum((state - goal) ** 2, axis=-1)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def _deceleration(self, state):
        center = self.deceleration["center"]
        decay = self.deceleration["decay"]

        center = tf.expand_dims(center, axis=0)
        decay = tf.expand_dims(decay, axis=0)
        state = tf.expand_dims(state, axis=1)

        delta = state - center
        delta = tf.squeeze(delta, axis=-1)
        distance = tf.norm(delta, axis=-1)

        lambdas = 2 / (1.0 + tf.exp(-decay * distance)) - 1.0
        return tf.reduce_prod(lambdas,  axis=-1)

    def __repr__(self):
        goal = self.goal.numpy().squeeze().tolist()

        low = self.action_space.low.squeeze().tolist()
        high = self.action_space.high.squeeze().tolist()
        bounds = f"[{low}, {high}]"

        center = self.deceleration["center"].numpy().tolist()
        decay = ", ".join([f"{d:.4f}" for d in self.deceleration["decay"].numpy().tolist()])
        deceleration = f"{{center={center}, decay=[{decay}]}}"

        return f"Navigation(goal={goal}, deceleration={deceleration}, bounds={bounds})"

    @classmethod
    def load(cls, config):
        goal = tf.constant(config["goal"])
        low = config["low"]
        high = config["high"]
        deceleration = {
            key: tf.constant(val)
            for key, val in config["deceleration"].items()
        }
        return cls(goal, deceleration, low, high)



import wandb


def setup(filename, filemode="w"):
    pass


def log(timestep, row):
    wandb.log(row, step=timestep)


def log_transition(timestep, state, action, cost, next_state, done, info):
    del next_state
    del done

    fluent_dict = {}
    for fluent_type, fluent in [("x", state), ("u", action)]:
        for i, value in enumerate(fluent):
            fluent_dict[f"{fluent_type}[{i}]"] = value

    total_cost = info["total_cost"]

    wandb.log(
        {
            **fluent_dict,
            "cost": cost,
            "reward": -cost,
            "cum_total_cost": total_cost,
            "cum_total_reward": -total_cost
        },
        step=timestep
    )


def summary(trajectory, uptime):
    total_cost = trajectory.total_cost
    wandb.run.summary["total_reward"] = -total_cost
    wandb.run.summary["total_cost"] = total_cost
    wandb.run.summary["uptime"] = uptime



import random

import numpy as np
import pytest
import tensorflow as tf

from tfmpc.envs.hvac import HVAC
from tfmpc.envs.navigation import Navigation
from tfmpc.envs.reservoir import Reservoir


@pytest.fixture
def batch_size():
    return random.randint(1, 32)


def sample_state(env, batch_size=1):
    state_size = env.state_size
    sample = [env.obs_space.sample() for _ in range(batch_size)]
    return tf.stack(sample)


def sample_action(env, batch_size=1):
    action_size = env.action_size
    sample = [env.action_space.sample() for _ in range(batch_size)]
    return tf.stack(sample)


def sample_adjacency_matrix(n):
    adj = (tf.random.uniform(shape=[n, n], maxval=1.0) >= 0.4)
    adj = tf.logical_and(adj, tf.eye(n) <= 0)
    adj = tf.linalg.band_part(adj, 0, -1) # upper triangular
    return adj


@pytest.fixture(scope="module", name="hvac")
def hvac_fixture():
    return hvac()


def hvac():
    n_rooms = tf.random.uniform(shape=[], minval=2, maxval=6, dtype=tf.int32)

    temp_outside = tf.random.normal(shape=[n_rooms, 1], mean=6.0, dtype=tf.float32)
    temp_hall = tf.random.normal(shape=[n_rooms, 1], mean=10.0, dtype=tf.float32)

    temp_lower_bound = tf.random.normal(shape=[n_rooms, 1], mean=20.0, stddev=1.5, dtype=tf.float32)
    temp_upper_bound = temp_lower_bound + tf.random.uniform(shape=[], minval=3.5, maxval=55.5)

    R_outside = tf.random.normal(shape=[n_rooms, 1], mean=4.0, dtype=tf.float32)
    R_hall = tf.random.normal(shape=[n_rooms, 1], mean=2.0, stddev=0.5, dtype=tf.float32)
    R_wall = tf.random.normal(shape=[n_rooms, n_rooms], mean=1.5, stddev=0.1, dtype=tf.float32)
    R_wall = 1 / 2 * (R_wall + tf.transpose(R_wall))

    capacity = tf.random.normal(shape=[n_rooms, 1], mean=80.0, stddev=2.0, dtype=tf.float32)

    air_max = tf.random.normal(shape=[n_rooms, 1], mean=15.0, dtype=tf.float32)

    adj = sample_adjacency_matrix(int(n_rooms))
    adj_outside = (tf.random.normal(shape=[n_rooms, 1]) >= 0.0)
    adj_hall = (tf.random.normal(shape=[n_rooms, 1]) >= 0.0)

    return HVAC(
        temp_outside, temp_hall,
        temp_lower_bound, temp_upper_bound,
        R_outside, R_hall, R_wall,
        capacity, air_max,
        adj, adj_outside, adj_hall
    )



MAX_RES_CAP = 100


def linear_topology(n):
    row = tf.constant([1] + [0] * (n - 1), dtype=tf.float32)
    rows = [tf.roll(row, shift=i, axis=0) for i in range(1, n)]
    rows.append(tf.zeros([n,], dtype=tf.float32))
    rows = tf.stack(rows, axis=0)
    return rows


@pytest.fixture(scope="module", name="reservoir")
def reservoir_fixture():
    return reservoir()


def reservoir():
    n_reservoirs = np.random.randint(low=2, high=5)

    max_res_cap = tf.constant(
        [MAX_RES_CAP] * n_reservoirs,
        shape=[n_reservoirs, 1], dtype=tf.float32)

    lower_bound = MAX_RES_CAP * tf.random.uniform(shape=[n_reservoirs, 1],
                                                  maxval=0.5,
                                                  dtype=tf.float32)
    upper_bound = MAX_RES_CAP * tf.random.uniform(shape=[n_reservoirs, 1],
                                                  minval=0.5,
                                                  dtype=tf.float32)

    downstream = linear_topology(n_reservoirs)

    rain_mean = 0.20 * MAX_RES_CAP
    rain_variance = 0.05 * (MAX_RES_CAP ** 2)
    rain_shape = tf.constant(
        [(rain_mean ** 2) / rain_variance] * n_reservoirs,
        shape=[n_reservoirs, 1],
        dtype=tf.float32)
    rain_scale = tf.constant(
        [rain_variance / rain_mean] * n_reservoirs,
        shape=[n_reservoirs, 1],
        dtype=tf.float32)

    low_penalty = tf.constant(
        [-5.0] * n_reservoirs,
        shape=[n_reservoirs, 1],
        dtype=tf.float32)
    high_penalty = tf.constant(
        [-100.0] * n_reservoirs,
        shape=[n_reservoirs, 1],
        dtype=tf.float32)
    set_point_penalty = tf.constant(
        [-0.1] * n_reservoirs,
        shape=[n_reservoirs, 1],
        dtype=tf.float32)

    return Reservoir(
        max_res_cap, lower_bound, upper_bound,
        low_penalty, high_penalty, set_point_penalty,
        downstream,
        rain_shape, rain_scale)


@pytest.fixture(params=[1, 2], ids=["1-zone", "2-zones"], name="navigation")
def navigation_fixture(request):
    yield navigation(request.param)


def navigation_1_zone():
    return navigation(1)


def navigation_2_zones():
    return navigation(2)


def navigation(n_zones):
    goal = tf.constant([[8.0], [9.0]])
    deceleration = {
        "center": tf.random.normal(shape=[n_zones, 2, 1]),
        "decay": tf.random.uniform(shape=[n_zones,], maxval=3.0)
    }
    low = [[-1.0], [-1.0]]
    high = [[1.0], [1.0]]
    return Navigation(goal, deceleration, low, high)



import pytest
import tensorflow as tf

from tfmpc.utils import optimization


@pytest.fixture(params=[
    ([0.0, 0.0], [-1.0, 0.5], [1.0, 1.0], [0.0, 0.5]),
    ([0.0, 0.0], [0.5, -1.0], [1.0, 1.0], [0.5, 0.0]),
    ([1.0, 1.0], [0.0, 1.5], [2.0, 2.0], [1.0, 1.5]),
    ([1.0, 1.0], [1.5, 0.0], [2.0, 2.0], [1.5, 1.0]),

    ([0.0, 0.0, 0.0], [-1.0, 0.5, -1.0], [1.0, 1.0, 1.0], [0.0, 0.5, 0.0]),
    ([0.0, 0.0, 0.0], [-1.0, 0.5, 0.30], [1.0, 1.0, 1.0], [0.0, 0.5, 0.30]),
])
def qp(request):
    goal, low, high, x_star = request.param

    dim = len(goal)

    H = 2 * tf.eye(dim)
    q = -2 * tf.constant(goal, shape=(dim, 1))
    low = tf.constant(low, shape=(dim, 1))
    high = tf.constant(high, shape=(dim, 1))
    x_star = tf.constant(x_star, shape=(dim, 1))

    return H, q, low, high, x_star


def test_get_qp_indices(qp):
    H, q, low, high, x_star = qp

    for _ in range(10):
        x = tf.random.uniform(
            tf.shape(x_star), minval=low + 1e-4, maxval=high - 1e-4)

        g = q + tf.matmul(H, x)
        f, c = optimization._get_qp_indices(g, low, high, x)

        assert tf.reduce_all(f == tf.constant([[True]] * len(x)))
        assert tf.reduce_all(c == tf.constant([[False]] * len(x)))


# def test_get_newton_step(qp):
#     H, q, low, high, x_star = qp

#     for _ in range(10):
#         x = tf.random.uniform(tf.shape(x_star), minval=low, maxval=high)

#         g = q + tf.matmul(H, x)
#         f, c = optimization._get_qp_indices(g, low, high, x)

#         delta_x = tf.Variable(tf.zeros_like(x))
#         H_ff, g_f = optimization._get_newton_step(H, q, x, f, c, delta_x)


def test_projected_newton_qp(qp):
    H, q, low, high, x_star = qp

    x, *_ = optimization.projected_newton_qp(H, q, low, high, x_star)
    assert tf.reduce_all(tf.abs(x - x_star) < 1e-4)

    for _ in range(10):
        x_0 = tf.random.uniform(tf.shape(x_star), minval=low, maxval=high)

        x, *_ = optimization.projected_newton_qp(H, q, low, high, x_0)
        assert tf.reduce_all(tf.abs(x - x_star) < 1e-4)

        for i in range(x_0.shape[0]):
            x = tf.Variable(x_0, trainable=False)
            x[i].assign(low[i])
            x, *_ = optimization.projected_newton_qp(H, q, low, high, x)
            assert tf.reduce_all(tf.abs(x - x_star) < 1e-4)

        for i in range(x_0.shape[0]):
            x = tf.Variable(x_0, trainable=False)
            x[i].assign(high[i])
            x, *_ = optimization.projected_newton_qp(H, q, low, high, x)
            assert tf.reduce_all(tf.abs(x - x_star) < 1e-4)



import numpy as np
import pytest
import tensorflow as tf

from tfmpc.utils.trajectory import Trajectory


@pytest.fixture
def trajectory():
    T = 10
    state_size = action_size = 3

    states = tf.random.normal(shape=(T + 1, state_size, 1))
    actions = tf.random.truncated_normal(shape=(T, action_size, 1))
    costs = tf.random.uniform(shape=(T,), minval=-1.0, maxval=1.0)

    return Trajectory(states, actions, costs)


def test_initial_state(trajectory):
    assert np.allclose(trajectory.initial_state, trajectory.states[0], atol=1e-3)


def test_final_state(trajectory):
    assert np.allclose(trajectory.final_state, trajectory.states[-1], atol=1e-3)


def test_total_cost(trajectory):
    assert np.allclose(trajectory.total_cost, sum(trajectory.costs), atol=1e-3)


def test_cumulative_cost(trajectory):
    costs = trajectory.costs
    expected = [np.sum(costs[: t + 1]) for t in range(len(costs))]
    assert np.allclose(trajectory.cumulative_cost, expected, atol=1e-3)


def test_cost_to_go(trajectory):
    costs = trajectory.costs
    expected = [np.sum(costs[t:]) for t in range(len(costs))]
    assert np.allclose(trajectory.cost_to_go, expected, atol=1e-3)


def test_len(trajectory):
    assert len(trajectory) == len(trajectory.actions)
    assert len(trajectory) == len(trajectory.costs)
    assert len(trajectory) == len(trajectory.states) - 1


def test_getitem(trajectory):
    for t, (x, u, c) in enumerate(trajectory):
        assert np.allclose(x, trajectory.states[t + 1], atol=1e-3)
        assert np.allclose(u, trajectory.actions[t], atol=1e-3)
        assert c == trajectory.costs[t]


def test_repr(trajectory):
    print()
    print(repr(trajectory))


def test_str(trajectory):
    print()
    print(str(trajectory))



# pylint: disable=missing-docstring

import numpy as np
import pytest
import tensorflow as tf

from tfmpc.envs.lqr.navigation import make_lqr_navigation_problem
from tfmpc.solvers.lqr import LQR


@pytest.fixture(scope="module")
def nav():
    goal = np.array([[8.32], [-5.5]])
    beta = 10.0
    return make_lqr_navigation_problem(goal, beta)


@pytest.fixture
def solver(nav):
    return LQR(nav)


def test_make_lqr_linear_navigation(nav):
    state_size = nav.state_size
    action_size = nav.action_size
    n_dim = state_size + action_size

    F = nav.F
    assert isinstance(F, tf.Variable)
    assert F.shape == (state_size, n_dim)

    f = nav.f
    assert isinstance(f, tf.Variable)
    assert f.shape == (state_size, 1)

    C = nav.C
    assert isinstance(C, tf.Variable)
    assert C.shape == (n_dim, n_dim)

    C = C.numpy()
    assert np.allclose(C.T, C, atol=1e-4)
    np.linalg.cholesky(C)
    assert np.all(np.linalg.eigvals(C) > 0)

    c = nav.c
    assert isinstance(c, tf.Variable)
    assert c.shape == (n_dim, 1)


def test_backward(solver):
    T = 10
    policy, value_fn = solver.backward(T)
    assert len(policy) == len(value_fn)


def test_forward(solver):
    nav = solver.lqr

    T = 10
    x0 = np.random.normal(size=(nav.state_size, 1)).astype("f")

    policy, _ = solver.backward(T)

    x, u, c = solver.forward(policy, x0, T)
    assert len(x) == len(u) + 1 == len(c)
    assert np.allclose(x[0], x0, atol=1e-4)

    F_t = nav.F.numpy()
    f_t = nav.f.numpy()
    C_t = nav.C.numpy()
    c_t = nav.c.numpy()

    for t in range(T):
        K, k = policy[t]
        action = np.dot(K, x[t]) + k
        assert np.allclose(action, u[t], atol=1e-4)

        inputs = np.concatenate([x[t], u[t]], axis=0)
        next_state = np.dot(F_t, inputs) + f_t
        assert np.allclose(next_state, x[t + 1], atol=1e-4)

        cost = 1 / 2 * np.dot(np.dot(inputs.T, C_t), inputs) + np.dot(c_t.T, inputs)
        assert np.allclose(cost, c[t], atol=1e-4)



from collections import namedtuple

import tensorflow as tf


TransitionApprox = namedtuple("TransitionApprox", "f f_x f_u")
CostApprox = namedtuple("CostApprox", "l l_x l_u l_xx l_uu l_ux l_xu")
FinalCostApprox = namedtuple("FinalCostApprox", "l l_x l_xx")


class DiffEnv:

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def get_linear_transition(self, state, action):
        with tf.GradientTape(persistent=True) as tape:
            tape.watch(state)
            tape.watch(action)
            f = self.transition(state, action)

        f_x = tape.batch_jacobian(f, state)
        f_u = tape.batch_jacobian(f, action)
        f_x = tf.squeeze(f_x, axis=[2, 4])
        f_u = tf.squeeze(f_u, axis=[2, 4])

        del tape

        return TransitionApprox(f, f_x, f_u)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def get_quadratic_cost(self, state, action):
        with tf.GradientTape(persistent=True) as tape:
            tape.watch(state)
            tape.watch(action)
            l = self.cost(state, action)
            l_x, l_u = tape.gradient(
                l, [state, action],
                unconnected_gradients=tf.UnconnectedGradients.ZERO)

        l_xx = tape.batch_jacobian(
            l_x, state,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)
        l_xu = tape.batch_jacobian(
            l_x, action,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)

        l_xx = tf.squeeze(l_xx, axis=[2, 4])
        l_xu = tf.squeeze(l_xu, axis=[2, 4])

        l_ux = tape.batch_jacobian(
            l_u, state,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)
        l_uu = tape.batch_jacobian(
            l_u, action,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)

        l_ux = tf.squeeze(l_ux, axis=[2, 4])
        l_uu = tf.squeeze(l_uu, axis=[2, 4])

        del tape

        return CostApprox(l, l_x, l_u, l_xx, l_uu, l_ux, l_xu)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, 1], dtype=tf.float32)
    ])
    def get_quadratic_final_cost(self, state):
        with tf.GradientTape(persistent=True) as tape:
            tape.watch(state)
            l = self.final_cost(state)
            l_x = tape.gradient(
                l, state,
                unconnected_gradients=tf.UnconnectedGradients.ZERO)

        l_xx = tape.jacobian(
            l_x, state,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)

        l_xx = tf.squeeze(l_xx, axis=[1, 3])

        del tape

        return FinalCostApprox(l, l_x, l_xx)



import pytest
import tensorflow as tf

from tests.conftest import sample_state, sample_action


def test_repr(reservoir):
    print(repr(reservoir))


def test_str(reservoir):
    print()
    print(str(reservoir))


def test_bounds(reservoir):
    assert reservoir.lower_bound.shape == [reservoir.state_size, 1]
    assert reservoir.upper_bound.shape == [reservoir.state_size, 1]
    assert tf.reduce_all(reservoir.lower_bound < reservoir.upper_bound)
    assert tf.reduce_all(reservoir.upper_bound < reservoir.max_res_cap)


def test_downstream(reservoir):
    assert reservoir.downstream.shape == [reservoir.state_size, reservoir.state_size]
    assert tf.reduce_all(tf.reduce_sum(reservoir.downstream, axis=1) >= 0)
    assert tf.reduce_all(tf.reduce_sum(reservoir.downstream, axis=1) <= 1)


def test_rain(reservoir):
    rain_mean = reservoir.rain_shape * reservoir.rain_scale
    assert tf.reduce_all(rain_mean >= 0.0)


def test_vaporated(reservoir):
    batch_size = 10

    rlevel1 = sample_state(reservoir)
    rlevel2 = sample_state(reservoir, batch_size=batch_size)

    for rlevel in [rlevel1, rlevel2]:
        batch_size = tf.shape(rlevel)[0]

        assert rlevel.shape == [batch_size, reservoir.state_size, 1]

        value = reservoir._vaporated(rlevel)
        assert value.shape == rlevel.shape
        assert tf.reduce_all(value > 0.0)

    fn = reservoir._vaporated
    cfn1 = fn.get_concrete_function(rlevel1)
    cfn2 = fn.get_concrete_function(rlevel2)
    assert cfn1 == cfn2


def test_inflow(reservoir):
    batch_size = 10

    outflow1 = sample_action(reservoir)
    outflow2 = sample_action(reservoir, batch_size=batch_size)

    for outflow in [outflow1, outflow2]:
        batch_size = tf.shape(outflow)[0]

        assert outflow.shape == [batch_size, reservoir.action_size, 1]
        assert tf.reduce_all(outflow >= 0.0)
        assert tf.reduce_all(outflow <= 1.0)

        value = reservoir._inflow(outflow)
        assert value.shape == [batch_size, reservoir.state_size, 1]

        assert tf.reduce_all(value[:, 0] == 0.0)
        assert tf.reduce_all(value[:, 1:] == outflow[:, :-1])

        balance = (
            tf.reduce_sum(value, axis=1)
            + outflow[:, -1]
            - tf.reduce_sum(outflow, axis=1)
        )
        assert tf.reduce_all(tf.abs(balance) < 1e-3)

    fn = reservoir._inflow
    assert fn.get_concrete_function(outflow1) == fn.get_concrete_function(outflow2)


def test_rainfall(reservoir):
    batch_size = 10

    rlevel1 = sample_action(reservoir)
    rlevel2 = sample_action(reservoir, batch_size=batch_size)

    for rlevel in [rlevel1, rlevel2]:
        batch_size = tf.shape(rlevel)[0]

        rainfall = reservoir._rainfall(batch_size)
        assert rainfall.shape == rlevel.shape

    fn = reservoir._rainfall
    batch_size1 = tf.shape(rlevel1)[0]
    batch_size2 = tf.shape(rlevel2)[0]
    cfn1 = fn.get_concrete_function(batch_size1)
    cfn2 = fn.get_concrete_function(batch_size2)
    assert cfn1 == cfn2


def test_outflow(reservoir):
    batch_size = 10

    flow1 = sample_action(reservoir)
    rlevel1 = sample_state(reservoir)

    flow2 = sample_action(reservoir, batch_size=batch_size)
    rlevel2 = sample_state(reservoir, batch_size=batch_size)

    for flow, rlevel in [(flow1, rlevel1), (flow2, rlevel2)]:
        assert flow.shape == rlevel.shape

        batch_size = tf.shape(rlevel)[0]

        outflow = reservoir._outflow(flow, rlevel)
        assert outflow.shape == [batch_size, reservoir.state_size, 1]

        assert tf.reduce_all(tf.abs(outflow / rlevel - flow) < 1e-3)

    fn = reservoir._outflow
    cfn1 = fn.get_concrete_function(flow1, rlevel1)
    cfn2 = fn.get_concrete_function(flow2, rlevel2)
    assert cfn1 == cfn2


def test_transition(reservoir):
    batch_size = 10

    state1 = sample_state(reservoir)
    action1 = sample_action(reservoir)

    state2 = sample_state(reservoir, batch_size=batch_size)
    action2 = sample_action(reservoir, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        assert state.shape == action.shape

        next_state = reservoir.transition(state, action)
        assert next_state.shape == state.shape

        rain_mean = reservoir.rain_shape * reservoir.rain_scale

        balance = (
            tf.reduce_sum(state, axis=1)
            + tf.expand_dims(tf.reduce_sum(rain_mean), axis=0)
            - tf.reduce_sum(reservoir._vaporated(state), axis=1)
            - action[:, -1] * state[:, -1]
            - tf.reduce_sum(next_state, axis=1)
        )
        assert tf.reduce_all(tf.abs(balance) < 1e-3)

    fn = reservoir.transition
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_linear_transition_model(reservoir):
    batch_size = 10

    state1 = sample_state(reservoir)
    action1 = sample_action(reservoir)

    state2 = sample_state(reservoir, batch_size=batch_size)
    action2 = sample_action(reservoir, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        assert state.shape == action.shape
        batch_size = tf.shape(state)[0]

        next_state = reservoir.transition(state, action)

        model = reservoir.get_linear_transition(state, action)
        f = model.f
        f_x = model.f_x
        f_u = model.f_u

        assert tf.reduce_all(tf.abs(f - next_state) < 1e-3)

        a_t = tf.squeeze(action, axis=-1)
        s_t = tf.squeeze(state, axis=-1)
        C = tf.squeeze(reservoir.max_res_cap, axis=-1)
        grad_v_s = tf.linalg.diag(1/2 * (tf.cos(s_t/C) * s_t/C + tf.sin(s_t/C)))
        grad_F_s = tf.linalg.diag(a_t)
        grad_I_s = tf.matmul(reservoir.downstream, tf.linalg.diag(a_t), transpose_a=True)
        f_x_expected = tf.eye(reservoir.state_size) - grad_v_s - grad_F_s + grad_I_s
        assert f_x.shape == f_x_expected.shape
        assert tf.reduce_all(tf.abs(f_x - f_x_expected) < 1e-3)

        grad_F_a = tf.linalg.diag(s_t)
        grad_I_a = tf.matmul(reservoir.downstream, tf.linalg.diag(s_t), transpose_a=True)
        f_u_expected = - grad_F_a + grad_I_a
        assert f_u.shape == f_u_expected.shape
        assert tf.reduce_all(tf.abs(f_u - f_u_expected) < 1e-3)

    fn = reservoir.get_linear_transition
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_cost(reservoir):
    batch_size = 10

    state1 = sample_state(reservoir)
    action1 = sample_action(reservoir)

    state2 = sample_state(reservoir, batch_size=batch_size)
    action2 = sample_action(reservoir, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        cost = reservoir.cost(state, action)
        assert cost.shape == [batch_size,]
        assert tf.reduce_all(cost >= 0.0)

    fn = reservoir.cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_final_cost(reservoir):
    state1 = sample_state(reservoir)[0]
    state2 = sample_state(reservoir)[0]

    for state in [state1, state2]:
        cost = reservoir.final_cost(state)
        assert cost.shape == []
        assert cost >= 0.0

    fn = reservoir.final_cost
    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_quadratic_cost_model(reservoir):
    batch_size = 10

    state1 = sample_state(reservoir)
    action1 = sample_action(reservoir)

    state2 = sample_state(reservoir, batch_size=batch_size)
    action2 = sample_action(reservoir, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        assert state.shape == action.shape
        batch_size = tf.shape(state)[0]

        cost = reservoir.cost(state, action)

        model = reservoir.get_quadratic_cost(state, action)
        l = model.l
        l_x = model.l_x
        l_u = model.l_u
        l_xx = model.l_xx
        l_uu = model.l_uu
        l_ux = model.l_ux
        l_xu = model.l_xu

        assert tf.reduce_all(tf.abs(l - cost) < 1e-3)

        s_t = tf.squeeze(state, axis=-1)
        LB = tf.squeeze(reservoir.lower_bound, axis=-1)
        UB = tf.squeeze(reservoir.upper_bound, axis=-1)
        HP = -tf.squeeze(reservoir.high_penalty, axis=-1)
        LP = -tf.squeeze(reservoir.low_penalty, axis=-1)
        SPP = -tf.squeeze(reservoir.set_point_penalty, axis=-1)

        e1 = HP * tf.cast(((s_t - UB) > 0.0), tf.float32)
        e2 = -LP * tf.cast(((LB - s_t) > 0.0), tf.float32)
        e3 = -SPP * tf.sign((UB + LB) / 2 - s_t)
        l_x_expected = tf.expand_dims(e1 + e2 + e3, -1)

        assert l_x.shape == state.shape
        assert tf.reduce_all(tf.abs(l_x - l_x_expected) < 1e-3)

        assert l_u.shape == action.shape
        assert tf.reduce_all(l_u == tf.zeros_like(action))

        assert l_xx.shape == [batch_size, reservoir.state_size, reservoir.state_size]
        assert tf.reduce_all(l_xx == tf.zeros([batch_size, reservoir.state_size, reservoir.state_size]))

        assert l_uu.shape == [batch_size, reservoir.action_size, reservoir.action_size]
        assert tf.reduce_all(l_uu == tf.zeros([batch_size, reservoir.action_size, reservoir.action_size]))

        assert l_ux.shape == [batch_size, reservoir.action_size, reservoir.state_size]
        assert tf.reduce_all(l_ux == tf.zeros([batch_size, reservoir.action_size, reservoir.state_size]))

        assert l_xu.shape == [batch_size, reservoir.state_size, reservoir.action_size]
        assert tf.reduce_all(l_xu == tf.zeros([batch_size, reservoir.state_size, reservoir.action_size]))

    fn = reservoir.get_quadratic_cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2



"""
Iterative Linear Quadratic Regulator (iLQR)

For details please see:

>> Control-Limited Differential Dynamic Programming (ICRA, 2014)
>> Tassa, Mansard, and Todorov (2014)
"""

import logging
import os
import time

import numpy as np
import tensorflow as tf
from tuneconfig import experiment

from tfmpc.envs import diffenv
from tfmpc.utils import optimization
from tfmpc.utils import trajectory
from tfmpc import loggers


class iLQR:

    def __init__(self, env, **kwargs):
        self.env = env

        # solve
        self.atol = kwargs.get("atol", 5e-3)
        self.max_iterations = kwargs.get("max_iterations", 100)

        # backward
        self.mu_min = kwargs.get("mu_min", 1e-6)
        self.delta_0 = kwargs.get("delta_0", 2.0)

        # forward
        self.c1 = kwargs.get("c1", 0.0)
        self.alpha_min = kwargs.get("alpha_min", 1e-3)

        # logging
        logdir = kwargs.get("logdir", "/tmp/ilqr")
        self.logger = kwargs.get("logger", loggers.get_logger("py_logging", logdir=logdir))

        self._config = kwargs

    @property
    def low(self):
        return tf.constant(self.env.action_space.low)

    @property
    def high(self):
        return tf.constant(self.env.action_space.high)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=None, dtype=tf.int32)
    ])
    def start(self, x0, T):
        states = tf.TensorArray(dtype=tf.float32, size=T+1)
        costs = tf.TensorArray(dtype=tf.float32, size=T+1)
        actions = tf.TensorArray(dtype=tf.float32, size=T)

        low = self.low
        minval = tf.where(tf.math.is_inf(low), -tf.ones_like(low), low)

        high = self.high
        maxval = tf.where(tf.math.is_inf(high), tf.ones_like(high), high)

        states = states.write(0, x0)

        state = x0

        for t in tf.range(T):
            action = tf.random.uniform(shape=[], minval=minval, maxval=maxval)

            state = tf.expand_dims(state, axis=0)
            action = tf.expand_dims(action, axis=0)

            cost = tf.squeeze(self.env.cost(state, action))
            state = self.env.transition(state, action)

            state = tf.reshape(state, tf.shape(x0))
            action = tf.squeeze(action, axis=0)

            actions = actions.write(t, action)
            costs = costs.write(t, cost)
            states = states.write(t+1, state)

        if self._config.get("ignore_final_cost"):
            final_cost = tf.zeros_like(costs.read(0))
        else:
            final_cost = tf.squeeze(self.env.final_cost(state))
        costs = costs.write(T, final_cost)

        return states.stack(), actions.stack(), costs.stack()

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def derivatives(self, states, actions):
        transition_model = self.env.get_linear_transition(states[:-1], actions)
        cost_model = self.env.get_quadratic_cost(states[:-1], actions)
        final_cost_model = self.env.get_quadratic_final_cost(states[-1])

        return transition_model, cost_model, final_cost_model

    @tf.function(input_signature=[
        tf.TensorSpec(shape=None, dtype=tf.int32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        diffenv.TransitionApprox(
            tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32)
        ),
        diffenv.CostApprox(
            tf.TensorSpec(shape=[None,], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None, None], dtype=tf.float32)
        ),
        diffenv.FinalCostApprox(
            tf.TensorSpec(shape=None, dtype=tf.float32),
            tf.TensorSpec(shape=[None, 1], dtype=tf.float32),
            tf.TensorSpec(shape=[None, None], dtype=tf.float32)
        ),
        tf.TensorSpec(shape=None, dtype=tf.float32)
    ])
    def backward(self, T, actions, transition_model, cost_model, final_cost_model, mu):
        state_size = self.env.state_size

        K = tf.TensorArray(dtype=tf.float32, size=T)
        k = tf.TensorArray(dtype=tf.float32, size=T)

        if self._config.get("ignore_final_cost"):
            V_x = tf.zeros([state_size, 1])
            V_xx = tf.zeros([state_size, state_size])
            J = 0.0
        else:
            V_x = final_cost_model.l_x
            V_xx = final_cost_model.l_xx
            J = final_cost_model.l

        dV1 = 0.0
        dV2 = 0.0

        for t in tf.range(T - 1, -1, -1):
            f_x = transition_model.f_x[t]
            f_u = transition_model.f_u[t]

            l = cost_model.l[t]
            l_x = cost_model.l_x[t]
            l_u = cost_model.l_u[t]
            l_xx = cost_model.l_xx[t]
            l_uu = cost_model.l_uu[t]
            l_xu = cost_model.l_xu[t]

            f_x_trans = tf.transpose(f_x)
            f_u_trans = tf.transpose(f_u)

            Q_x = l_x + tf.matmul(f_x_trans, V_x)
            Q_u = l_u + tf.matmul(f_u_trans, V_x)

            f_x_trans_V_xx = tf.matmul(f_x_trans, V_xx)
            f_u_trans_V_xx = tf.matmul(f_u_trans, V_xx)
            f_u_trans_V_xx_reg = tf.matmul(f_u_trans, V_xx + mu * tf.eye(state_size))

            Q_xx = l_xx + tf.matmul(f_x_trans_V_xx, f_x)
            Q_uu = l_uu + tf.matmul(f_u_trans_V_xx, f_u)
            Q_ux = tf.transpose(l_xu) + tf.matmul(f_u_trans_V_xx, f_x)

            Q_uu_reg = l_uu + tf.matmul(f_u_trans_V_xx_reg, f_u)
            Q_ux_reg = tf.transpose(l_xu) + tf.matmul(f_u_trans_V_xx_reg, f_x)

            if self.env.action_space.is_bounded():
                if tf.math.count_nonzero(V_xx) > 0:
                    K_t, k_t = tf.py_function(self._get_constrained_controller, inp=[actions[t], Q_uu_reg, Q_ux_reg, Q_u], Tout=[tf.float32, tf.float32])
                else:
                    K_t = tf.zeros([self.env.action_size, self.env.state_size])
                    k_t = tf.where(Q_u >= 0.0, self.env.action_space.low - actions[t], self.env.action_space.high - actions[t])
            else:
                K_t, k_t = self._get_unconstrained_controller(Q_uu_reg, Q_ux_reg, Q_u)

            K_t_trans = tf.transpose(K_t)
            k_t_trans = tf.transpose(k_t)
            K_t_trans_Q_uu = tf.matmul(K_t_trans, Q_uu)

            V_x = tf.reshape(
                (Q_x
                 + tf.matmul(Q_ux, k_t, transpose_a=True)
                 + tf.matmul(K_t_trans, Q_u)
                 + tf.matmul(K_t_trans_Q_uu, k_t)),
                [self.env.state_size, 1])

            V_xx = tf.reshape(
                (Q_xx
                 + tf.matmul(Q_ux, K_t, transpose_a=True)
                 + tf.matmul(K_t_trans, Q_ux)
                 + tf.matmul(K_t_trans_Q_uu, K_t)),
                [self.env.state_size, self.env.state_size])
            V_xx = 1 / 2 * (V_xx + tf.transpose(V_xx))

            J += l

            dV1 += tf.reshape(tf.squeeze(tf.matmul(k_t_trans, Q_u)), [])
            dV2 += tf.reshape(1 / 2 * tf.squeeze(tf.matmul(tf.matmul(k_t_trans, Q_uu), k_t)), [])

            K = K.write(t, K_t)
            k = k.write(t, k_t)

        return K.stack(), k.stack(), J, dV1, dV2

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, None], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=None, dtype=tf.float32)
    ])
    def forward(self, x, u, K, k, alpha):
        T = tf.shape(x)[0] - 1

        states = tf.TensorArray(size=T+1, dtype=tf.float32)
        costs = tf.TensorArray(size=T+1, dtype=tf.float32)
        actions = tf.TensorArray(size=T, dtype=tf.float32)

        x0 = x[0]
        u0 = u[0]

        states = states.write(0, x0)

        J = tf.constant(0.0)

        state = x0
        residual = tf.constant(0.0)

        low = self.low
        high = self.high

        for t in tf.range(T):
            tf.autograph.experimental.set_loop_options(
                shape_invariants=[(J, tf.TensorShape(None))]
            )

            delta_x = state - x[t]
            delta_u = alpha * k[t] + tf.matmul(K[t], delta_x)

            action = u[t] + delta_u
            action = tf.clip_by_value(action, low, high)

            state = tf.expand_dims(state, axis=0)
            action = tf.expand_dims(action, axis=0)
            cost = tf.squeeze(self.env.cost(state, action))
            state = self.env.transition(state, action)
            state = tf.reshape(state, tf.shape(x0))
            action = tf.reshape(action, tf.shape(u0))

            actions = actions.write(t, action)
            costs = costs.write(t, cost)
            states = states.write(t + 1, state)

            J += cost
            residual = tf.math.maximum(residual, tf.reduce_max(tf.abs(delta_u)))

        if self._config.get("ignore_final_cost"):
            final_cost = tf.zeros_like(costs.read(0))
        else:
            final_cost = tf.squeeze(self.env.final_cost(state))
        costs = costs.write(T, final_cost)
        J += final_cost

        return states.stack(), actions.stack(), costs.stack(), J, residual

    def solve(self, x0, T, show_progress=True):
        mu = 0.0
        delta = 1.0

        x_hat, u_hat, c_hat = self.start(x0, T)

        epochs = self.max_iterations
        run_id = self._config.get("run_id", 0)
        num_workers = self._config.get("num_workers", 1)

        show_progress = show_progress or self._config.get("show_progress", False)

        desc = f"horizon={int(T)}"
        with experiment.Experiment.trange(epochs, run_id, num_workers, desc=desc, show_progress=show_progress) as t:

            for iteration in t:

                logging.info(f"[SOLVE] >>>>>>> Iteration = {iteration} <<<<<<<")

                start = time.time()

                # model approximation
                transition_model, cost_model, final_cost_model = self.derivatives(x_hat, u_hat)

                converged = False

                while True:
                    # backward pass
                    K, k, J_hat, dV1, dV2 = self._backward(T, u_hat, transition_model, cost_model, final_cost_model, mu, delta)

                    # check for termination due to small gradient
                    g_norm = tf.reduce_mean(tf.reduce_max(tf.abs(k) / (tf.abs(u_hat) + 1.0),axis=1), axis=0)[0]
                    logging.debug(f"[SOLVE] g_norm = {g_norm:.6f}")
                    if g_norm < self.atol:
                        logging.debug(f"[SOLVE] CONVERGED: g_norm < atol ({g_norm:.4f} < {self.atol})")
                        converged = True
                        break

                    # forward pass (backtracking line-search)
                    done, x, u, c, residual = self._forward(x_hat, u_hat, J_hat, K, k, dV1, dV2)

                    if residual < self.atol:
                        logging.debug(f"[SOLVE] CONVERGED: residual < atol ({residual:.6f} < {self.atol:.6f})")
                        converged = True
                        x_hat, u_hat, c_hat = x, u, c
                        break

                    if done:
                        # decrease regularization
                        delta = min(1 / self.delta_0, delta / self.delta_0)
                        mu = mu * delta * (mu * delta > self.mu_min)

                        # accept improvement
                        x_hat, u_hat, c_hat = x, u, c
                        break
                    else:
                        # increase regularization
                        delta = max(self.delta_0, delta * self.delta_0)
                        mu = max(self.mu_min, mu * delta)

                uptime = time.time() - start
                logging.info(f"[SOLVE] uptime = {uptime:.4f} sec")

                # convergence test
                if converged:
                    break

                t.set_postfix(J=f"{J_hat:.4f}", g_norm=f"{g_norm:.4f}", residual=f"{residual:.4f}")

        traj = trajectory.Trajectory(x_hat, u_hat, c_hat)

        return traj, iteration

    def _backward(self, T, u_hat, transition_model, cost_model, final_cost_model, mu, delta):
        T = tf.constant(T, dtype=tf.int32)

        done = False
        num_iter = 0

        while not done:
            num_iter += 1
            logging.debug(f"[BACKWARD] num_iter = {num_iter}, mu = {mu}, delta = {delta}")

            try:
                start = time.time()
                mu_ = tf.constant(mu, dtype=tf.float32)
                K, k, J_hat, dV1, dV2 = self.backward(T, u_hat, transition_model, cost_model, final_cost_model, mu_)
                uptime = time.time() - start

                done = True

                logging.info(f"[BACKWARD] num_iter = {num_iter}, uptime = {uptime:.4f} sec")
                logging.info(f"[BACKWARD] num_iter = {num_iter}, J_hat = {J_hat:.4f}")

            except tf.errors.InvalidArgumentError as e:
                logging.warn(f"[BACKWARD] could not run iLQR.backward : {e}")

                delta = max(self.delta_0, delta * self.delta_0)
                mu = max(self.mu_min, mu * delta)

            except Exception as e:
                logging.fatal(f"[BACKWARD] Error: {e}")
                exit(-1)

        return K, k, J_hat, dV1, dV2

    def _forward(self, x_hat, u_hat, J_hat, K, k, dV1, dV2):
        accept = False

        num_iter = 0

        for alpha in np.geomspace(1.0, self.alpha_min, 11):
            num_iter += 1

            logging.debug(f"[FORWARD] num_iter = {num_iter}, alpha = {alpha}")

            start = time.time()
            alpha_ = tf.constant(alpha, dtype=tf.float32)
            x, u, c, J, residual = self.forward(x_hat, u_hat, K, k, alpha_)
            uptime = time.time() - start

            logging.info(f"[FORWARD] num_iter = {num_iter}, uptime = {uptime:.4f} sec, J = {J:.4f}")
            logging.debug(f"[FORWARD] num_iter = {num_iter}, residual = {residual}")

            # if residual < self.atol:
            #     accept = True
            #     break

            delta_J = - alpha * (dV1 + alpha * dV2)
            dcost = J_hat - J

            if delta_J > 0:
                z = dcost / delta_J
            else:
                z = tf.sign(dcost)
                logging.warn(f"[FORWARD] Non-positive expected reduction: delta_J = {delta_J:.4f}")

            logging.debug(f"[FORWARD] num_iter = {num_iter}, dcost = {dcost}, delta_J = {delta_J:.4f}")
            logging.debug(f"[FORWARD] num_iter = {num_iter}, z = {z}, c1 = {self.c1}")

            if z >= self.c1:
                accept = True
                break

        return accept, x, u, c, residual

    def _get_unconstrained_controller(self, Q_uu, Q_ux, Q_u):
        R = tf.linalg.cholesky(Q_uu)
        kK = -tf.linalg.cholesky_solve(R, tf.concat([Q_u, Q_ux], axis=1))
        k = kK[:,:1]
        K = kK[:,1:]
        return K, k

    def _get_constrained_controller(self, u, Q_uu, Q_ux, Q_u):
        low = self.env.action_space.low - u
        high = self.env.action_space.high - u

        #k_0 = tf.zeros_like(u)
        k_0 = (low + high) / 2

        k, Hfree, free, clamped = optimization.projected_newton_qp(Q_uu, Q_u, low, high, k_0)

        action_dim = self.env.action_size
        state_dim = self.env.state_size
        K = tf.Variable(tf.zeros([action_dim, state_dim]))

        n_free = tf.math.count_nonzero(tf.squeeze(free))
        if n_free > 0:
            free = tf.squeeze(free)
            Q_ux_f = Q_ux[free]
            indices = tf.where(free)
            values = - tf.linalg.cholesky_solve(Hfree, Q_ux_f)
            K.scatter_nd_update(indices, values)

        K = tf.constant(K.numpy())

        return K, k



import contextlib
import time

import tensorflow as tf

from tfmpc.utils.trajectory import Trajectory


class Runner:

    def __init__(self, env, agent, **kwargs):
        self.env = env
        self.agent = agent

        self._on_step_hook = kwargs.get("on_step")
        self._on_episode_end_hook = kwargs.get("on_episode_end")

    def _on_step(self, timestep, state, action, cost, next_state, done, info):
        if self._on_step_hook:
            self._on_step_hook(timestep, state, action, cost, next_state, done, info)

    def _on_episode_end(self, trajectory, uptime):
        if self._on_episode_end:
            self._on_episode_end_hook(trajectory, uptime)

    def run(self, mode=None):
        start_time = time.perf_counter()

        state = self.env.reset()
        timestep = 0
        done = False

        states = [state]
        actions = []
        costs = []

        total_cost = 0.0

        while not done:
            action = self.agent(state, timestep)
            next_state, cost, done, info = self.env.step(action)

            total_cost += cost

            self._on_step(timestep, state, action, cost, next_state, done, info)

            if mode is not None:
                self.env.render(mode)

            state = next_state
            timestep = self.env._t

            states.append(state)
            actions.append(action)
            costs.append(cost)

        costs.append(self.env.final_cost(state))

        states = tf.stack(states)
        actions = tf.stack(actions)
        costs = tf.stack(costs)

        trajectory = Trajectory(states, actions, costs)
        uptime = time.perf_counter() - start_time

        self._on_episode_end(trajectory, uptime)

        return trajectory

    @contextlib.contextmanager
    def __call__(self, initial_state, horizon):
        self.env.setup(initial_state, horizon)
        yield self
        self.env.close()



# pylint: disable=missing-docstring

import numpy as np
import pytest
import tensorflow as tf

from tfmpc.envs.lqr import make_random_lqr_problem
from tfmpc.solvers.lqr import LQR


@pytest.fixture(scope="module")
def lqr():
    state_size = np.random.randint(2, 10)
    action_size = np.random.randint(2, 10)
    return make_random_lqr_problem(state_size, action_size)


@pytest.fixture
def solver(lqr):
    return LQR(lqr)

def test_make_lqr(lqr):
    state_size = lqr.state_size
    action_size = lqr.action_size
    n_dim = state_size + action_size

    F = lqr.F
    assert isinstance(F, tf.Variable)
    assert F.shape == (state_size, n_dim)

    f = lqr.f
    assert isinstance(f, tf.Variable)
    assert f.shape == (state_size, 1)

    C = lqr.C
    assert isinstance(C, tf.Variable)
    assert C.shape == (n_dim, n_dim)

    C = C.numpy()
    assert np.allclose(C.T, C, atol=1e-2)
    np.linalg.cholesky(C)
    assert np.all(np.linalg.eigvals(C) > 0)

    c = lqr.c
    assert isinstance(c, tf.Variable)
    assert c.shape == (n_dim, 1)


def test_backward(solver):
    T = 10
    policy, value_fn = solver.backward(T)
    assert len(policy) == len(value_fn) == T


def test_forward(solver):
    lqr = solver.lqr

    T = 10
    x0 = np.random.normal(size=(lqr.state_size, 1)).astype("f")

    policy, value_fn = solver.backward(T)

    x, u, c = solver.forward(policy, x0, T)
    assert len(x) == len(u) + 1 == len(c)
    assert np.allclose(x[0], x0, atol=1e-2)

    F_t = lqr.F.numpy()
    f_t = lqr.f.numpy()
    C_t = lqr.C.numpy()
    c_t = lqr.c.numpy()

    for t in range(T):
        K, k = policy[t]
        action = np.dot(K, x[t]) + k
        assert np.allclose(action, u[t], atol=1e-2)

        inputs = np.concatenate([x[t], u[t]], axis=0)
        next_state = np.dot(F_t, inputs) + f_t
        assert np.allclose(next_state, x[t + 1], atol=1e-2)

        cost = 1 / 2 * np.dot(np.dot(inputs.T, C_t), inputs) + np.dot(c_t.T, inputs)
        assert np.allclose(cost, c[t], atol=1e-2)

        V, v, const = value_fn[t]
        V = V.numpy()
        v = v.numpy()
        const = const.numpy()

        x_t = x[t].numpy()
        value = const + 1 / 2 * np.dot(np.dot(x_t.T, V), x_t) + np.dot(v.T, x_t)
        cost_to_go = np.sum(c[t:])
        assert np.allclose(value, cost_to_go, atol=1e-2)


def test_solve(lqr, solver):
    x0 = np.random.normal(size=(lqr.state_size, 1)).astype("f")
    T = 10
    trajectory = solver.solve(x0, T)
    assert len(trajectory.states) == len(trajectory.actions) + 1
    assert len(trajectory.states) == len(trajectory.costs)


def test_dump_and_load(lqr):
    tmpfile = "/tmp/config.json"

    with open(tmpfile, "w") as file:
        lqr.dump(file)

    with open(tmpfile, "r") as file:
        lqr2 = lqr.load(file)

    for k in lqr.__dict__:
        assert tf.reduce_all(getattr(lqr, k) == getattr(lqr2, k))



import importlib


def make_env(config):
    module = config["module"]
    module = importlib.import_module(f"tfmpc.envs.{module}")
    cls_name = config["cls_name"]
    return getattr(module, cls_name).load(config["config"])



import pytest
import tensorflow as tf
import tensorflow.compat.v1.logging as tf_logging

from tfmpc.envs.lqr import navigation
from tfmpc.solvers import ilqr

tf_logging.set_verbosity(tf_logging.ERROR)


@pytest.fixture(params=[[[5.5], [-9.0]]])
def goal(request):
    return request.param


@pytest.fixture(params=[0.0, 5.0], ids=["beta=0.0", "beta=5.0"])
def beta(request):
    return request.param


@pytest.fixture(params=[None, [-1.0, 1.0]], ids=["unconstrained", "box"])
def bounds(request):
    return request.param


@pytest.fixture
def initial_state():
    return tf.constant([[0.0], [0.0]])


@pytest.fixture
def horizon():
    return tf.constant(10)


@pytest.fixture(scope="function")
def solver(goal, beta, bounds):
    goal = tf.constant(goal)
    beta = tf.constant(beta)
    low = high = None
    if bounds:
        low, high = bounds

    env = navigation.NavigationLQR(goal, beta, low, high)
    return ilqr.iLQR(env)


def test_start(solver, initial_state, horizon):
    x0 = initial_state
    T = horizon
    states, actions, costs = solver.start(x0, T)

    assert isinstance(states, tf.Tensor)
    assert states.shape == [T + 1, *x0.shape]
    assert tf.reduce_all(states[0] == x0)

    assert isinstance(actions, tf.Tensor)
    assert actions.shape == [T, *x0.shape]

    for t in range(T):
        x, u = states[t], actions[t]
        assert tf.reduce_all(states[t + 1] == solver.env.transition(x, u))


def test_derivatives(solver, initial_state, horizon):
    states, actions, costs = solver.start(initial_state, horizon)
    models = solver.derivatives(states, actions)

    assert len(models) == 3
    transition_model, cost_model, final_cost_model = models

    assert all(tf.shape(g)[0] == horizon for g in transition_model)
    assert all(tf.shape(g)[0] == horizon for g in cost_model)
    #assert all(tf.shape(g)[0] == horizon for g in final_cost_model)


def test_backward(solver, initial_state, horizon):
    states, actions, costs = solver.start(initial_state, horizon)
    transition_model, cost_model, final_cost_model = solver.derivatives(states, actions)
    mu = 1.0
    K, k, J, dV1, dV2 = solver.backward(horizon, actions, transition_model, cost_model, final_cost_model, mu)

    action_size = solver.env.action_size
    state_size = solver.env.state_size

    assert isinstance(K, tf.Tensor)
    assert K.shape == [horizon, action_size, state_size]

    assert isinstance(k, tf.Tensor)
    assert k.shape == [horizon, action_size, 1]


def test_forward(solver, initial_state, horizon):
    x, u, c = solver.start(initial_state, horizon)
    transition_model, cost_model, final_cost_model = solver.derivatives(x, u)
    mu = 1.0
    alpha = 1.0
    K, k, J, dV1, dV2 = solver.backward(horizon, u, transition_model, cost_model, final_cost_model, mu)
    states, actions, costs, J_hat, residual = solver.forward(x, u, K, k, alpha)

    assert states.shape == x.shape
    assert actions.shape == u.shape
    assert costs.shape == [horizon + 1]

    assert tf.reduce_all(x[0] == states[0])

    for t in range(horizon):
        x, u, c = states[t], actions[t], costs[t]
        assert tf.reduce_all(states[t + 1] == solver.env.transition(x, u))
        assert c == solver.env.cost(x, u)

    assert costs[horizon] == solver.env.final_cost(states[horizon])


def test_solve(solver, initial_state, horizon):
    trajectory, iterations = solver.solve(initial_state, horizon, show_progress=False)
    #assert iterations == 2
    print()
    print(trajectory)



import random

import numpy as np
import pytest
import tensorflow as tf

from tests.conftest import sample_state, sample_action
from tfmpc.envs.lqr import navigation


@pytest.fixture(scope="module")
def env():
    goal = tf.constant([[8.0], [-9.0]])
    beta = 5.0
    return navigation.NavigationLQR(goal, beta)


def test_transition(env, batch_size):
    for _ in range(5):
        x0 = sample_state(env, batch_size)
        u0 = sample_action(env, batch_size)

        x1 = env.transition(x0, u0)

        assert isinstance(x1, tf.Tensor)
        assert x1.shape == x0.shape
        assert tf.reduce_all(tf.abs(x1 - (x0 + u0)) < 1e-4)


def test_linear_transition(env, batch_size):
    for _ in range(5):
        x = sample_state(env, batch_size)
        u = sample_action(env, batch_size)

        model = env.get_linear_transition(x, u)
        f = model.f
        f_x = model.f_x
        f_u = model.f_u

        assert isinstance(f, tf.Tensor)
        assert f.shape == x.shape
        assert tf.reduce_all(f == env.transition(x, u))

        assert isinstance(f_x, tf.Tensor)
        assert f_x.shape == [batch_size] + [env.state_size] * 2
        assert tf.reduce_all(f_x == tf.eye(2))

        assert isinstance(f_u, tf.Tensor)
        assert f_u.shape == [batch_size] + [env.action_size] * 2
        assert tf.reduce_all(f_u == tf.eye(2))


def test_cost(env, batch_size):
    for _ in range(5):
        x = sample_state(env, batch_size)
        u = sample_action(env, batch_size)
        c = env.cost(x, u)

        assert isinstance(c, tf.Tensor)
        assert c.shape == [batch_size,]


def test_final_cost(env, batch_size):
    for _ in range(5):
        x = sample_state(env, batch_size)
        c = env.final_cost(x)

        assert isinstance(c, tf.Tensor)
        assert c.shape == [batch_size,]


def test_quadratic_cost(env, batch_size):
    goal = env.goal
    beta = env.beta

    for _ in range(5):
        x = sample_state(env, batch_size)
        u = sample_action(env, batch_size)

        model = env.get_quadratic_cost(x, u)
        l = model.l
        l_x = model.l_x
        l_u = model.l_u
        l_xx = model.l_xx
        l_uu = model.l_uu
        l_xu = model.l_xu
        l_ux = model.l_ux

        assert isinstance(l, tf.Tensor)
        assert l.shape == [batch_size,]
        assert tf.reduce_all(l == env.cost(x, u))

        assert isinstance(l_x, tf.Tensor)
        assert l_x.shape == x.shape
        assert tf.reduce_all(l_x == 2 * (x - goal))

        assert isinstance(l_u, tf.Tensor)
        assert l_u.shape == u.shape
        assert tf.reduce_all(l_u == 2 * beta * u)

        assert isinstance(l_xx, tf.Tensor)
        assert l_xx.shape == [batch_size] + [env.state_size] * 2
        assert tf.reduce_all(l_xx == 2 * tf.eye(2))

        assert isinstance(l_uu, tf.Tensor)
        assert l_uu.shape == [batch_size] + [env.action_size] * 2
        assert tf.reduce_all(l_uu == 2 * beta * tf.eye(2))

        assert isinstance(l_ux, tf.Tensor)
        assert l_ux.shape == (batch_size, env.state_size, env.action_size)
        assert tf.reduce_all(l_ux == tf.zeros((env.state_size, env.action_size)))

        assert isinstance(l_xu, tf.Tensor)
        assert l_xu.shape == (batch_size, env.action_size, env.state_size)
        assert tf.reduce_all(l_xu == tf.zeros((env.action_size, env.state_size)))


def test_quadratic_final_cost(env):
    goal = env.goal

    for _ in range(5):
        x = env.obs_space.sample()

        model = env.get_quadratic_final_cost(x)
        l = model.l
        l_x = model.l_x
        l_xx = model.l_xx

        assert isinstance(l, tf.Tensor)
        assert l.shape == []
        assert tf.reduce_all(l == env.final_cost(x))

        assert isinstance(l_x, tf.Tensor)
        assert l_x.shape == x.shape
        assert tf.reduce_all(l_x == 2 * (x - goal))

        assert isinstance(l_xx, tf.Tensor)
        assert l_xx.shape == [env.state_size] * 2
        assert tf.reduce_all(l_xx == 2 * tf.eye(2))



import gym
import numpy as np
import tensorflow as tf

from tfmpc.envs.diffenv import DiffEnv
from tfmpc.envs.gymenv import GymEnv


class HVAC(DiffEnv, GymEnv):

    CAP_AIR = tf.constant(1.006, dtype=tf.float32)
    COST_AIR = tf.constant(1.0, dtype=tf.float32)
    TEMP_AIR = tf.constant(40.0, dtype=tf.float32)
    TIME_DELTA = tf.constant(1.0, dtype=tf.float32)
    PENALTY = tf.constant(20_000, dtype=tf.float32)
    SET_POINT_PENALTY = tf.constant(10.0, dtype=tf.float32)

    def __init__(self,
                 temp_outside,
                 temp_hall,
                 temp_lower_bound,
                 temp_upper_bound,
                 R_outside,
                 R_hall,
                 R_wall,
                 capacity,
                 air_max,
                 adj,
                 adj_outside,
                 adj_hall):

        super().__init__()

        # temperature outside and hall
        self.temp_outside = temp_outside
        self.temp_hall = temp_hall

        # temperature limits
        self.temp_lower_bound = temp_lower_bound
        self.temp_upper_bound = temp_upper_bound

        # thermal conductance
        self.R_outside = R_outside
        self.R_hall = R_hall
        self.R_wall = R_wall

        # heat capacity
        self.capacity = capacity

        # air maximum volume
        self.air_max = air_max

        # thermal conductance noise
        self.air_std = 1.0
        self.room_std = 1.0
        self.outside_std = 1.0
        self.hall_std = 1.0

        # room adjacency matrix
        self.adj = adj
        self.adj_outside = adj_outside
        self.adj_hall = adj_hall

        self.obs_space = gym.spaces.Box(
            shape=[self.state_size, 1], low=-np.inf, high=np.inf)

        self.action_space = gym.spaces.Box(
            shape=[self.action_size, 1], low=0.0, high=1.0)

    @property
    def state_size(self):
        return len(self.temp_lower_bound)

    @property
    def action_size(self):
        return self.state_size

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def transition(self, state, action):
        temp = state
        air = action * self.air_max

        heating = self._heating(air, temp)

        conduction_between_rooms = self._conduction_between_rooms(temp)
        conduction_with_outside = self._conduction_with_outside(temp)
        conduction_with_hall = self._conduction_with_hall(temp)

        temp_ = (
            temp
            + self.TIME_DELTA / self.capacity * (
                heating
                + conduction_between_rooms
                + conduction_with_outside
                + conduction_with_hall
            )
        )
        return temp_

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def cost(self, state, action):
        temp = state
        air = action * self.air_max

        air_cost = self.COST_AIR * air
        out_of_bounds_penalty = self.PENALTY * (
            tf.maximum(0.0, self.temp_lower_bound - temp)
            + tf.maximum(0.0, temp - self.temp_upper_bound)
        )
        set_point_penalty = (
            self.SET_POINT_PENALTY
            * tf.abs(
                (self.temp_lower_bound + self.temp_upper_bound) / 2 - temp)
        )

        return tf.reduce_sum(
            tf.squeeze(air_cost + out_of_bounds_penalty + set_point_penalty, axis=-1),
            axis=-1
        )

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, 1], dtype=tf.float32)
    ])
    def final_cost(self, state):
        temp = state

        out_of_bounds_penalty = self.PENALTY * (
            tf.maximum(0.0, self.temp_lower_bound - temp)
            + tf.maximum(0.0, temp - self.temp_upper_bound)
        )
        set_point_penalty = (
            self.SET_POINT_PENALTY
            * tf.abs(
                (self.temp_lower_bound + self.temp_upper_bound) / 2 - temp)
        )

        return tf.reduce_sum(
            tf.squeeze(out_of_bounds_penalty + set_point_penalty, axis=-1),
            axis=-1
        )

    @tf.function
    def _heating(self, air, temp):
        return air * self.CAP_AIR * (self.TEMP_AIR - temp)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def _conduction_between_rooms(self, temp):
        adj = tf.logical_or(self.adj, tf.transpose(self.adj))
        adj = tf.cast(adj, tf.float32)
        return tf.reduce_sum(
            - adj / self.R_wall * (temp - tf.linalg.matrix_transpose(temp)),
            axis=-1,
            keepdims=True
        )

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def _conduction_with_outside(self, temp):
        adj_outside = tf.cast(self.adj_outside, tf.float32)
        return adj_outside / self.R_outside * (self.temp_outside - temp)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def _conduction_with_hall(self, temp):
        adj_hall = tf.cast(self.adj_hall, tf.float32)
        return adj_hall / self.R_hall * (self.temp_hall - temp)

    def __repr__(self):
        return f"HVAC({self.state_size})"

    def __str__(self):
        bounds = [
            f"[{float(low):.3f}, {float(high):.3f}]"
            for low, high in zip(
                    tf.squeeze(self.temp_lower_bound),
                    tf.squeeze(self.temp_upper_bound))]
        bounds = f"[{', '.join(bounds)}]"

        R = [
            f"[outside={float(o):.3f}, hall={float(h):.3f}]"
            for o, h in zip(
                    tf.squeeze(self.R_outside),
                    tf.squeeze(self.R_hall))]
        R = "\n".join(R)

        R_wall = self.R_wall.numpy()

        capacity = [f"{c:.3f}" for c in tf.squeeze(self.capacity)]
        capacity = f"[{', '.join(capacity)}]"

        air_max = [f"{c:.3f}" for c in tf.squeeze(self.air_max)]
        air_max = f"[{', '.join(air_max)}]"

        adj = self.adj.numpy()
        adj_outside = tf.squeeze(self.adj_outside).numpy().tolist()
        adj_hall = tf.squeeze(self.adj_hall).numpy().tolist()

        temp_outside = [f"{c:.3f}" for c in tf.squeeze(self.temp_outside)]
        temp_outside = f"[{', '.join(temp_outside)}]"

        temp_hall = [f"{c:.3f}" for c in tf.squeeze(self.temp_hall)]
        temp_hall = f"[{', '.join(temp_hall)}]"

        return f"HVAC(\ntemp_bounds={bounds},\nR=\n{R},\nR_wall=\n{R_wall},\ncapacity={capacity},\nair_max={air_max},\nadj=\n{adj},\nadj_outside={adj_outside},\nadj_hall={adj_hall},\ntemp_outside={temp_outside},\ntemp_hall={temp_hall}\n)"
    @classmethod
    def load(cls, config):
        for key, value in config.items():
            if key.startswith("adj"):
                config[key] = tf.constant(value, dtype=tf.bool)
            else:
                config[key] = tf.constant(value, dtype=tf.float32)
        return cls(**config)



import json

import numpy as np
import tensorflow as tf

from tfmpc.utils import trajectory


class LQR:

    def __init__(self, lqr):
        self.lqr = lqr

    @tf.function
    def backward(self, T):
        policy, value_fn = [], []

        state_size = self.lqr.state_size

        F, f, C, c = self.lqr.F, self.lqr.f, self.lqr.C, self.lqr.c

        V = C[:state_size, :state_size]
        v = c[:state_size]
        const = 0.0

        value_fn.append((V, v, const))

        for t in reversed(range(T)):
            F_trans_V = tf.matmul(tf.transpose(F), V)
            Q = C + tf.matmul(F_trans_V, F)
            q = (c
                 + tf.matmul(F_trans_V, f)
                 + tf.matmul(tf.transpose(F), v))

            Q_uu = Q[state_size:, state_size:]
            Q_ux = Q[state_size:, :state_size]
            q_u = q[state_size:]

            inv_Q_uu = tf.linalg.inv(Q_uu)

            K = -tf.matmul(inv_Q_uu, Q_ux)
            k = -tf.matmul(inv_Q_uu, q_u)

            K_trans = tf.transpose(K)
            k_trans = tf.transpose(k)

            Q_xx = Q[:state_size, :state_size]
            Q_xu = Q[:state_size, state_size:]
            q_x = q[:state_size]
            K_Q_uu = tf.matmul(K_trans, Q_uu)

            V = (Q_xx
                 + tf.matmul(Q_xu, K)
                 + tf.matmul(K_trans, Q_ux)
                 + tf.matmul(K_Q_uu, K))

            v = (q_x
                 + tf.matmul(Q_xu, k)
                 + tf.matmul(K_trans, q_u)
                 + tf.matmul(K_Q_uu, k))

            V_, v_, _ = value_fn[-1]

            F_trans = tf.transpose(F)
            f_trans = tf.transpose(f)
            V_f = tf.matmul(V_, f)

            W = C + tf.matmul(F_trans, tf.matmul(V_, F))
            w = c + tf.matmul(F_trans, V_f) + tf.matmul(F_trans, v_)
            W_uu = W[state_size:, state_size:]
            w_u = w[state_size:]

            const1 = 1 / 2 * tf.matmul(k_trans, tf.matmul(W_uu, k))
            const2 = tf.matmul(k_trans, w_u)
            const3 = 1/ 2 * tf.matmul(f_trans, V_f) + tf.matmul(f_trans, v_)
            const += (const1 + const2 + const3)

            policy.append((K, k))
            value_fn.append((V, v, const))

        policy = list(reversed(policy))
        value_fn = list(reversed(value_fn[1:]))

        return policy, value_fn

    @tf.function
    def forward(self, policy, x0, T):
        states = [x0]
        actions = []
        costs = []

        state = x0

        for t in range(T):
            K, k = policy[t]
            action = tf.matmul(K, state) + k

            next_state = self.lqr.transition(state, action)
            cost = self.lqr.cost(state, action)

            state = next_state

            states.append(next_state)
            actions.append(action)
            costs.append(cost)

        final_cost = self.lqr.final_cost(state)
        costs.append(final_cost)

        states = tf.stack(states, axis=0)
        actions = tf.stack(actions, axis=0)
        costs = tf.stack(costs, axis=0)

        return states, actions, costs

    def solve(self, x0, T):
        policy, value_fn = self.backward(T)
        states, actions, costs = self.forward(policy, x0, T)
        return trajectory.Trajectory(states, actions, costs)



import numpy as np
import pytest
import tensorflow as tf

from tests.conftest import sample_state, sample_action


def test_str(hvac):
    print()
    print(hvac)


def test_repr(hvac):
    print()
    print(repr(hvac))


def test_temp_bounds(hvac):
    assert hvac.temp_lower_bound.shape == [hvac.state_size, 1]
    assert hvac.temp_upper_bound.shape == [hvac.state_size, 1]
    assert tf.reduce_all(hvac.temp_lower_bound <= hvac.temp_upper_bound)


def test_thermal_conductance(hvac):
    assert hvac.R_outside.shape == [hvac.state_size, 1]
    assert hvac.R_hall.shape == [hvac.state_size, 1]
    assert hvac.R_wall.shape == [hvac.state_size, hvac.state_size]

    assert tf.reduce_all(hvac.R_outside > 0.0)
    assert tf.reduce_all(hvac.R_hall > 0.0)
    assert tf.reduce_all(hvac.R_wall > 0.0)


def test_adj(hvac):
    adj = hvac.adj
    assert adj.shape == [hvac.state_size, hvac.state_size]
    assert tf.reduce_all(tf.linalg.diag_part(adj) == False)
    assert tf.reduce_all(tf.linalg.band_part(adj, -1, 0) == False)


def test_transition(hvac):
    batch_size = 10

    state1 = sample_state(hvac)
    action1 = sample_action(hvac)

    state2 = sample_state(hvac, batch_size=batch_size)
    action2 = sample_action(hvac, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        next_state = hvac.transition(state, action)
        assert next_state.shape == state.shape

    fn = hvac.transition

    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_linear_transition(hvac):
    state1 = sample_state(hvac)
    action1 = sample_action(hvac)

    batch_size = 10
    state2 = sample_state(hvac, batch_size=batch_size)
    action2 = sample_action(hvac, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:

        model = hvac.get_linear_transition(state, action)

        # check f
        next_state = hvac.transition(state, action)
        assert model.f.shape == next_state.shape
        assert tf.reduce_all(model.f == next_state)

        # check f_x
        air = action * hvac.air_max
        heating_x = - tf.linalg.diag(tf.squeeze(air * hvac.CAP_AIR, axis=-1))

        adj = tf.logical_or(hvac.adj, tf.transpose(hvac.adj))
        adj = tf.cast(adj, tf.float32)
        conduction_between_rooms_x = - adj / hvac.R_wall * (-tf.ones_like(adj))

        adj_outside = tf.cast(hvac.adj_outside, tf.float32)
        conduction_with_outside_x = - tf.linalg.diag(
            tf.squeeze(adj_outside / hvac.R_outside))

        adj_hall = tf.cast(hvac.adj_hall, tf.float32)
        conduction_with_hall_x = - tf.linalg.diag(
            tf.squeeze(adj_hall / hvac.R_hall))

        C = tf.linalg.diag(tf.squeeze(hvac.TIME_DELTA / hvac.capacity))
        expected_f_x = (
            tf.eye(hvac.state_size)
            + tf.matmul(C,
                heating_x
                + conduction_between_rooms_x
                + conduction_with_outside_x
                + conduction_with_hall_x
            )
        )

        # print(expected_f_x)
        # print(model.f_x)
        assert expected_f_x.shape == model.f_x.shape
        # assert tf.reduce_all(tf.abs(model.f_x - expected_f_x) < 1e-3)

        # check f_u
        temp = state
        expected_f_u = tf.linalg.diag(
            tf.squeeze(
                hvac.TIME_DELTA / hvac.capacity * hvac.air_max * hvac.CAP_AIR * (hvac.TEMP_AIR - temp),
                axis=-1))
        assert model.f_u.shape == expected_f_u.shape
        assert tf.reduce_all(tf.abs(model.f_u - expected_f_u) < 1e-3)

    fn = hvac.get_linear_transition
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_conduction_between_rooms(hvac):
    state1 = sample_state(hvac)
    state2 = sample_state(hvac, batch_size=10)

    for state in [state1, state2]:
        batch_size = tf.shape(state)[0]
        state_size = tf.shape(state)[1]

        conduction = hvac._conduction_between_rooms(state)
        assert conduction.shape == state.shape

        adj = tf.logical_or(hvac.adj, tf.transpose(hvac.adj))
        R_wall = hvac.R_wall
        temp = tf.squeeze(state, axis=-1)

        expected = np.empty(shape=[batch_size, state_size], dtype=np.float32)
        for i in range(batch_size):
            for s in range(state_size):
                c = 0.0
                for p in range(state_size):
                    c += float(adj[s, p]) * (temp[i, p] - temp[i, s]) / R_wall[s, p]
                expected[i, s] = c
        expected = tf.constant(expected,
                               shape=[batch_size, state_size, 1],
                               dtype=tf.float32)

        assert expected.shape == conduction.shape
        assert tf.reduce_all(tf.abs(conduction - expected) < 1e-3)

    fn = hvac._conduction_between_rooms

    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_conduction_with_outside(hvac):
    state1 = sample_state(hvac)
    state2 = sample_state(hvac, batch_size=10)

    for state in [state1, state2]:
        batch_size = tf.shape(state)[0]

        conduction = hvac._conduction_with_outside(state)
        assert conduction.shape == state.shape

        for i in range(batch_size):
            conduction_ = conduction[i]
            state_ = state[i]

            idx1 = tf.where(tf.squeeze(hvac.adj_outside))
            cond = tf.gather(tf.squeeze(conduction_), idx1)
            temp_diff = tf.gather(tf.squeeze(hvac.temp_outside - state_), idx1)
            assert tf.reduce_all(tf.sign(cond) == tf.sign(temp_diff))

            idx2 = tf.where(tf.squeeze(tf.logical_not(hvac.adj_outside)))
            cond = tf.gather(tf.squeeze(conduction_), idx2)
            assert tf.reduce_all(cond == 0.0)

    fn = hvac._conduction_with_outside

    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_conduction_with_hall(hvac):
    state1 = sample_state(hvac)
    state2 = sample_state(hvac, batch_size=10)

    for state in [state1, state2]:
        batch_size = tf.shape(state)[0]

        conduction = hvac._conduction_with_hall(state)
        assert conduction.shape == state.shape

        for i in range(batch_size):
            conduction_ = conduction[i]
            state_ = state[i]

            idx1 = tf.where(tf.squeeze(hvac.adj_hall))
            cond = tf.gather(tf.squeeze(conduction_), idx1)
            temp_diff = tf.gather(tf.squeeze(hvac.temp_hall - state_), idx1)
            assert tf.reduce_all(tf.sign(cond) == tf.sign(temp_diff))

            idx2 = tf.where(tf.squeeze(tf.logical_not(hvac.adj_hall)))
            cond = tf.gather(tf.squeeze(conduction_), idx2)
            assert tf.reduce_all(cond == 0.0)

    fn = hvac._conduction_with_hall

    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_cost(hvac):
    state1 = sample_state(hvac)
    action1 = sample_action(hvac)

    batch_size = 10
    state2 = sample_state(hvac, batch_size=batch_size)
    action2 = sample_action(hvac, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        cost = hvac.cost(state, action)
        assert cost.shape == [batch_size,]
        assert tf.reduce_all(cost >= 0.0)

    fn = hvac.cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2


def test_final_cost(hvac):
    state1 = sample_state(hvac)[0]
    state2 = sample_state(hvac)[0]

    for state in [state1, state2]:
        cost = hvac.final_cost(state)
        assert cost.shape == []
        assert cost >= 0.0

    fn = hvac.final_cost
    cfn1 = fn.get_concrete_function(state1)
    cfn2 = fn.get_concrete_function(state2)
    assert cfn1 == cfn2


def test_quadratic_cost(hvac):
    state1 = sample_state(hvac)
    action1 = sample_action(hvac)

    batch_size = 10
    state2 = sample_state(hvac, batch_size=batch_size)
    action2 = sample_action(hvac, batch_size=batch_size)

    for state, action in [(state1, action1), (state2, action2)]:
        batch_size = tf.shape(state)[0]

        model = hvac.get_quadratic_cost(state, action)

        # check l
        assert tf.reduce_all(model.l == hvac.cost(state, action))

        # check l_x
        temp = state

        below_limit = - tf.linalg.diag(
            tf.squeeze(tf.sign(tf.maximum(0.0, hvac.temp_lower_bound - temp)),
                       axis=-1))
        above_limit = tf.linalg.diag(
            tf.squeeze(tf.sign(tf.maximum(0.0, temp - hvac.temp_upper_bound)),
                       axis=-1))
        out_of_bounds_penalty_x = (
            hvac.PENALTY * (below_limit + above_limit)
        )

        set_point_penalty_x = - hvac.SET_POINT_PENALTY * tf.linalg.diag(
            tf.squeeze(tf.sign((hvac.temp_lower_bound + hvac.temp_upper_bound) / 2 - temp),
                       axis=-1))

        expected_l_x = tf.reduce_sum(
            out_of_bounds_penalty_x + set_point_penalty_x,
            axis=-1,
            keepdims=True)

        assert model.l_x.shape == expected_l_x.shape
        assert tf.reduce_all(tf.abs(model.l_x - expected_l_x) < 1e-3)

        # check l_u
        expected_l_u = tf.reshape(
            tf.tile(hvac.air_max * hvac.COST_AIR, [batch_size, 1]),
            shape=[batch_size, -1, 1])
        assert model.l_u.shape == expected_l_u.shape
        assert tf.reduce_all(tf.abs(model.l_u - expected_l_u) < 1e-3)

        # check all 2nd order derivatives are null
        assert tf.reduce_all(model.l_xx == tf.zeros([hvac.state_size, hvac.state_size]))
        assert tf.reduce_all(model.l_uu == tf.zeros([hvac.state_size, hvac.state_size]))
        assert tf.reduce_all(model.l_xu == tf.zeros([hvac.state_size, hvac.state_size]))
        assert tf.reduce_all(model.l_ux == tf.zeros([hvac.state_size, hvac.state_size]))

    fn = hvac.get_quadratic_cost
    cfn1 = fn.get_concrete_function(state1, action1)
    cfn2 = fn.get_concrete_function(state2, action2)
    assert cfn1 == cfn2



import json

import numpy as np
from sklearn.datasets import make_spd_matrix
import tensorflow as tf


def make_random_lqr_problem(state_size, action_size):
    n_dim = state_size + action_size

    F = np.random.normal(size=(state_size, n_dim))
    f = np.random.normal(size=(state_size, 1))

    C = make_spd_matrix(n_dim)
    c = np.random.normal(size=(n_dim, 1))

    return LQR(F, f, C, c)


class LQR:
    """
    Linear Quadratic Regulator (LQR)

    For notation and more details on LQR please check out:
    - http://rail.eecs.berkeley.edu/deeprlcourse/static/slides/lec-10.pdf
    """

    def __init__(self, F, f, C, c):
        self.F = tf.Variable(F, trainable=False, dtype=tf.float32, name="F")
        self.f = tf.Variable(f, trainable=False, dtype=tf.float32, name="f")
        self.C = tf.Variable(C, trainable=False, dtype=tf.float32, name="C")
        self.c = tf.Variable(c, trainable=False, dtype=tf.float32, name="c")

    @property
    def n_dim(self):
        return self.F.shape[1]

    @property
    def state_size(self):
        return self.F.shape[0]

    @property
    def action_size(self):
        return self.n_dim - self.state_size

    @tf.function
    def transition(self, x, u):
        inputs = tf.concat([x, u], axis=0)
        return tf.matmul(self.F, inputs) + self.f

    @tf.function
    def cost(self, x, u):
        inputs = tf.concat([x, u], axis=0)
        inputs_transposed = tf.transpose(inputs)
        c1 = 1 / 2 * tf.matmul(tf.matmul(inputs_transposed, self.C), inputs)
        c2 = tf.matmul(inputs_transposed, self.c)
        return tf.squeeze(c1 + c2)

    @tf.function
    def final_cost(self, x):
        state_size = self.state_size
        C_xx = self.C[:state_size, :state_size]
        c_x = self.c[:state_size]
        x_transposed = tf.transpose(x)
        c1 = 1 / 2 * tf.matmul(tf.matmul(x_transposed, C_xx), x)
        c2 = tf.matmul(x_transposed, c_x)
        return tf.squeeze(c1 + c2)

    def dump(self, file):
        config = {
            "F": self.F.numpy().tolist(),
            "f": self.f.numpy().tolist(),
            "C": self.C.numpy().tolist(),
            "c": self.c.numpy().tolist()
        }
        json.dump(config, file)

    @classmethod
    def load(cls, file):
        config = json.load(file)
        config = {k: np.array(v).astype("f") for k, v in config.items()}
        return cls(**config)



import gym
import numpy as np
import tensorflow as tf

from tfmpc.envs.diffenv import DiffEnv
from tfmpc.envs.gymenv import GymEnv
from tfmpc.envs.lqr import LQR


def make_lqr_navigation_problem(goal, beta):
    state_size = action_size = goal.shape[0]

    F = np.concatenate([np.identity(state_size)] * action_size, axis=1).astype("f")
    f = np.zeros((state_size, 1)).astype("f")

    C = np.diag([2.0] * state_size + [2.0 * beta] * action_size).astype("f")
    c = np.concatenate([-2.0 * goal, np.zeros((action_size, 1))], axis=0).astype("f")

    return LQR(F, f, C, c)


class NavigationLQR(DiffEnv, GymEnv):

    def __init__(self, goal, beta, low=None, high=None):
        self.goal = goal
        self.beta = beta

        if low is None:
            low = -np.inf
        if high is None:
            high = np.inf

        self.obs_space = gym.spaces.Box(-np.inf, np.inf, shape=tf.shape(goal))
        self.action_space = gym.spaces.Box(low, high, shape=tf.shape(goal))

    @property
    def action_size(self):
        return self.state_size

    @property
    def state_size(self):
        return self.goal.shape[0]

    @tf.function
    def transition(self, state, action):
        return state + action

    @tf.function
    def cost(self, state, action):
        state = tf.squeeze(state, axis=-1)
        goal = tf.squeeze(self.goal, axis=-1)
        action = tf.squeeze(action)
        c1 = tf.reduce_sum((state - goal) ** 2, axis=-1)
        c2 = tf.reduce_sum(action ** 2, axis=-1)
        return c1 + self.beta * c2

    @tf.function
    def final_cost(self, state):
        state = tf.squeeze(state, axis=-1)
        goal = tf.squeeze(self.goal, axis=-1)
        return tf.reduce_sum((state - goal) ** 2, axis=-1)

    @classmethod
    def load(cls, config):
        goal = config["goal"]
        state_dim = len(goal)
        goal = tf.constant(goal, shape=(state_dim, 1), dtype=tf.float32)
        beta = tf.constant(config["beta"], dtype=tf.float32)
        low = config.get("low")
        high = config.get("high")

        return cls(goal, beta, low, high)

    def __repr__(self):
        goal = self.goal.numpy().squeeze().tolist()
        beta = self.beta
        bounds = ""
        if self.action_space.is_bounded():
            low = self.action_space.low.squeeze().tolist()
            high = self.action_space.high.squeeze().tolist()
            bounds = f", bounds=[{low}, {high}]"
        return f"NavigationLQR(goal={goal}, beta={beta}{bounds})"



import numpy as np
import tensorflow as tf

class MPC:

    def __init__(self, solver, horizon):
        self.solver = solver
        self.horizon = horizon

    def __call__(self, state, timestep):
        steps_to_go = self.horizon - timestep

        trajectory, iterations = self.solver.solve(state, steps_to_go)
        self.solver.logger.log(timestep, {"iterations": iterations})

        action = trajectory[0].action
        action = tf.constant(action[..., np.newaxis], dtype=tf.float32)
        return action



import gym.core
import tensorflow as tf


class GymEnv(gym.core.Env):

    def __init__(self):
        self._t = None
        self._state = None
        self._info = {}

    def setup(self, initial_state, horizon):
        self.initial_state = initial_state
        self.horizon = horizon

    def step(self, action):
        self._t += 1

        state = tf.expand_dims(self._state, axis=0)
        action = tf.expand_dims(action, axis=0)

        next_state = tf.squeeze(self.transition(state, action), axis=0)
        cost = tf.squeeze(self.cost(state, action))
        done = (self._t == self.horizon)

        self._info["total_cost"] += cost
        info = self._info

        self._state = next_state

        return (next_state, cost, done, info)

    def reset(self):
        self._t = 0
        self._state = self.initial_state
        self._info = {
            "total_cost": 0.0
        }

        return self._state

    def render(self, mode="human"):
        pass

    def close(self):
        pass

    def seed(self, seed=None):
        pass



import importlib
import os


def get_logger(module, logdir=None):
    logger = importlib.import_module(f"tfmpc.loggers.{module}")

    if logdir:
        filename = os.path.join(logdir, "trace.log")
        logger.setup(filename=filename)

    return logger



import json
import os

import tensorflow as tf

from tfmpc import agents
from tfmpc import envs
from tfmpc import runners
from tfmpc.solvers import ilqr
from tfmpc.loggers import get_logger


def ilqr_run(config):
    env_config = config.pop("env")
    with open(env_config, "r") as file:
        env_config = json.load(file)

    env = envs.make_env(env_config)

    x0 = tf.constant(env_config["initial_state"], dtype=tf.float32)
    T = tf.constant(config.pop("horizon"), dtype=tf.int32)

    solver = ilqr.iLQR(env, **config)
    trajectory, iterations = solver.solve(x0, T)

    output = os.path.join(config["logdir"], "data.csv")
    trajectory.save(output)

    return env, trajectory


def online_ilqr_run(config):
    import os

    import psutil

    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    os.environ["OMP_NUM_THREADS"] = str(psutil.cpu_count(logical=False))

    env_config = config.pop("env")
    with open(env_config, "r") as file:
        env_config = json.load(file)

    env = envs.make_env(env_config)

    x0 = tf.constant(env_config["initial_state"], dtype=tf.float32)
    T = tf.constant(config.pop("horizon"), dtype=tf.int32)

    logger = get_logger(config.get("logger", "py_logging"), config["logdir"])
    config["logger"] = logger

    solver = ilqr.iLQR(env, **config)
    controller = agents.MPC(solver, T)

    runner = runners.Runner(env, controller,
                            on_step=logger.log_transition,
                            on_episode_end=logger.summary)

    with runner(x0, T) as r:
        trajectory = r.run()
        output = os.path.join(config["logdir"], "data.csv")
        trajectory.save(output)

    return env, trajectory




import logging


def setup(filename, filemode="w"):
    logging.basicConfig(filename=filename, filemode=filemode, level=logging.DEBUG)


def log(timestep, row):
    logging.info(f">> timestep = {timestep} : {row}")


def log_transition(timestep, state, action, cost, next_state, done, info):
    logging.info(f">> timestep = {timestep}")
    logging.info(f" state = {state.numpy().tolist()}")
    logging.info(f" action = {action.numpy().tolist()}")
    logging.info(f" cost = {cost}")
    logging.info(f" cumulative_total_cost = {info['total_cost']}")


def summary(trajectory, uptime):
    total_cost = trajectory.total_cost
    logging.info(f">> total_cost = {total_cost}")
    logging.info(f">> uptime = {uptime}")



import gym
import numpy as np
import tensorflow as tf

from tfmpc.envs.diffenv import DiffEnv
from tfmpc.envs.gymenv import GymEnv


class Reservoir(DiffEnv, GymEnv):

    def __init__(self,
                 max_res_cap,
                 lower_bound,
                 upper_bound,
                 low_penalty,
                 high_penalty,
                 set_point_penalty,
                 downstream,
                 rain_shape,
                 rain_scale):
        super().__init__()

        self.max_res_cap = max_res_cap
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.low_penalty = low_penalty
        self.high_penalty = high_penalty
        self.set_point_penalty = set_point_penalty
        self.downstream = downstream
        self.rain_shape = rain_shape
        self.rain_scale = rain_scale

        self.obs_space = gym.spaces.Box(
            low=np.zeros_like(self.max_res_cap), high=np.array(self.max_res_cap))

        self.action_space = gym.spaces.Box(
            shape=[self.action_size, 1], low=0.0, high=1.0)

    @property
    def state_size(self):
        return len(self.lower_bound)

    @property
    def action_size(self):
        return self.state_size

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def transition(self, state, action):
        rlevel = state
        batch_size = tf.shape(rlevel)[0]

        outflow = self._outflow(action, state)
        inflow = self._inflow(outflow)

        vaporated = self._vaporated(state)
        rainfall = self._rainfall(batch_size)

        rlevel_ = (
            rlevel
            + rainfall + inflow
            - vaporated - outflow
        )
        return rlevel_

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def cost(self, state, action):
        return self._rlevel_penalty_cost(state)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, 1], dtype=tf.float32)
    ])
    def final_cost(self, state):
        state = tf.expand_dims(state, axis=0)
        return tf.squeeze(self._rlevel_penalty_cost(state))

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def _vaporated(self, rlevel):
        return (1.0 / 2.0) * tf.sin(rlevel / self.max_res_cap) * rlevel

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def _inflow(self, outflow):
        return tf.matmul(self.downstream, outflow, transpose_a=True)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32),
    ])
    def _outflow(self, relative_flow, rlevel):
        return relative_flow * rlevel

    @tf.function
    def _rainfall(self, batch_size):
        return tf.reshape(
            tf.tile(self.rain_shape * self.rain_scale, [batch_size, 1]),
            shape=[batch_size, -1, 1])

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, 1], dtype=tf.float32)
    ])
    def _rlevel_penalty_cost(self, rlevel):
        low = self.lower_bound
        high = self.upper_bound

        c1 = -self.low_penalty * tf.maximum(0.0, low - rlevel)
        c2 = -self.high_penalty * tf.maximum(0.0, rlevel - high)
        c3 = -self.set_point_penalty * tf.abs((low + high) / 2.0 - rlevel)

        penalty = tf.reduce_sum(tf.squeeze(c1 + c2 + c3, axis=-1), axis=-1)
        return penalty

    def __repr__(self):
        return f"Reservoir({self.state_size})"

    def __str__(self):
        bounds = ", ".join(
            f"[{float(l):.2f}, {float(u):.2f}]"
            for l, u in zip(self.lower_bound, self.upper_bound))

        topology = self.downstream

        rain = ", ".join(
            f"Gamma(shape={float(shape):.2f}, scale={float(scale):.2f})"
            for shape, scale in zip(self.rain_shape, self.rain_scale))

        return f"Reservoir(\nbounds={bounds},\ntopology=\n{topology},\nrain={rain})"

    @classmethod
    def load(cls, config):
        kwargs = {
            key: tf.constant(val, dtype=tf.float32)
            for key, val in config.items()
        }
        return cls(**kwargs)
